# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import numpy.testing as nptest
import numpy as np
from pygeogrids.geodetic_datum import GeodeticDatum


class test_GeodaticDatum(unittest.TestCase):

    def setUp(self):
        self.datum = GeodeticDatum('WGS84')

    def test_toECEF(self):
        x, y, z = self.datum.toECEF(0, 90)
        nptest.assert_almost_equal(np.array([0, 0, self.datum.geod.b]),
                                   np.array([x, y, z]),
                                   decimal=5)

        x, y, z = self.datum.toECEF(0, 0)
        nptest.assert_almost_equal(np.array([self.datum.geod.a, 0, 0]),
                                   np.array([x, y, z]),
                                   decimal=5)

    def test_ParallelRadi(self):
        r = self.datum.ParallelRadi(0.)
        nptest.assert_almost_equal(r, self.datum.geod.a, decimal=5)

        r = self.datum.ParallelRadi(90.)
        nptest.assert_almost_equal(r, 0., decimal=5)

    def test_GeocentricDistance(self):
        r = self.datum.GeocentricDistance(0., 0.)
        nptest.assert_almost_equal(r, self.datum.geod.a, decimal=5)

        r = self.datum.GeocentricDistance(0., 90.)
        nptest.assert_almost_equal(r, self.datum.geod.b, decimal=5)

    def test_ParallelArcDist(self):
        dist = self.datum.ParallelArcDist(0., 0., 360.)
        nptest.assert_almost_equal(dist, self.datum.geod.a * 2 * np.pi)

        lat, lon1, lon2 = 45., -5., 5.
        __, __, great_circle_dist = self.datum.geod.inv(lon1, lat, lon2, lat)
        parallel_dist = self.datum.ParallelArcDist(lat, lon1, lon2)

        assert great_circle_dist < parallel_dist, \
            (great_circle_dist, parallel_dist)


if __name__ == "__main__":
    unittest.main()



#! -*- coding: utf-8 -*-

"""
    Retrieval of version number

    This file helps to compute a version number in source trees obtained from
    git-archive tarball (such as those provided by githubs download-from-tag
    feature). Distribution tarballs (built by setup.py sdist) and build
    directories (produced by setup.py build) will contain a much shorter file
    that just contains the computed version number.
    This file was generated by PyScaffold.
"""

import inspect
import os
import re
import subprocess
import sys

__location__ = os.path.join(os.getcwd(), os.path.dirname(
    inspect.getfile(inspect.currentframe())))

# these strings will be replaced by git during git-archive
git_refnames = "$Format:%d$"
git_full = "$Format:%H$"

# general settings
tag_prefix = 'v'  # tags are like v1.2.0
package = "pygeogrids"
namespace = []
root_pkg = namespace[0] if namespace else package
if namespace:
    pkg_path = os.path.join(*namespace[-1].split('.') + [package])
else:
    pkg_path = package


class ShellCommand(object):
    def __init__(self, command, shell=True, cwd=None):
        self._command = command
        self._shell = shell
        self._cwd = cwd

    def __call__(self, *args):
        command = "{cmd} {args}".format(cmd=self._command,
                                        args=subprocess.list2cmdline(args))
        output = subprocess.check_output(command,
                                         shell=self._shell,
                                         cwd=self._cwd,
                                         stderr=subprocess.STDOUT,
                                         universal_newlines=True)
        return self._yield_output(output)

    def _yield_output(self, msg):
        for line in msg.splitlines():
            yield line


def get_git_cmd(**args):
    if sys.platform == "win32":
        for cmd in ["git.cmd", "git.exe"]:
            git = ShellCommand(cmd, **args)
            try:
                git("--version")
            except (subprocess.CalledProcessError, OSError):
                continue
            return git
        return None
    else:
        git = ShellCommand("git", **args)
        try:
            git("--version")
        except (subprocess.CalledProcessError, OSError):
            return None
        return git


def version_from_git(tag_prefix, root, verbose=False):
    # this runs 'git' from the root of the source tree. This only gets called
    # if the git-archive 'subst' keywords were *not* expanded, and
    # _version.py hasn't already been rewritten with a short version string,
    # meaning we're inside a checked out source tree.
    git = get_git_cmd(cwd=root)
    if not git:
        print("no git found")
        return None
    try:
        tag = next(git("describe", "--tags", "--dirty", "--always"))
    except subprocess.CalledProcessError:
        return None
    if not tag.startswith(tag_prefix):
        if verbose:
            print("tag '{}' doesn't start with prefix '{}'".format(tag,
                                                                   tag_prefix))
        return None
    tag = tag[len(tag_prefix):]
    sha1 = next(git("rev-parse", "HEAD"))
    full = sha1.strip()
    if tag.endswith("-dirty"):
        full += "-dirty"
    return {"version": tag, "full": full}


def get_keywords(versionfile_abs):
    # the code embedded in _version.py can just fetch the value of these
    # keywords. When used from setup.py, we don't want to import _version.py,
    # so we do it with a regexp instead. This function is not used from
    # _version.py.
    keywords = dict()
    try:
        with open(versionfile_abs, "r") as fh:
            for line in fh.readlines():
                if line.strip().startswith("git_refnames ="):
                    mo = re.search(r'=\s*"(.*)"', line)
                    if mo:
                        keywords["refnames"] = mo.group(1)
                if line.strip().startswith("git_full ="):
                    mo = re.search(r'=\s*"(.*)"', line)
                    if mo:
                        keywords["full"] = mo.group(1)
    except EnvironmentError:
        return None
    return keywords


def version_from_keywords(keywords, tag_prefix, verbose=False):
    if not keywords:
        return None  # keyword-finding function failed to find keywords
    refnames = keywords["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("keywords are unexpanded, not using")
        return None  # unexpanded, so not in an unpacked git-archive tarball
    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set([r[len(TAG):] for r in refs if r.startswith(TAG)])
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set([r for r in refs if re.search(r'\d', r)])
        if verbose:
            print("discarding '{}', no digits".format(",".join(refs-tags)))
    if verbose:
        print("likely tags: {}".format(",".join(sorted(tags))))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix):]
            if verbose:
                print("picking {}".format(r))
            return {"version": r,
                    "full": keywords["full"].strip()}
    else:
        if verbose:
            print("no suitable tags, using full revision id")
        return {"version": keywords["full"].strip(),
                "full": keywords["full"].strip()}


def version_from_parentdir(parentdir_prefix, root, verbose=False):
    # Source tarballs conventionally unpack into a directory that includes
    # both the project name and a version string.
    dirname = os.path.basename(root)
    if not dirname.startswith(parentdir_prefix):
        if verbose:
            print("guessing rootdir is '{}', but '{}' doesn't start with "
                  "prefix '{}'".format(root, dirname, parentdir_prefix))
        return None
    version = dirname[len(parentdir_prefix):].split('-')[0]
    return {"version": version, "full": ""}


def git2pep440(ver_str):
    dash_count = ver_str.count('-')
    if dash_count == 0:
        return ver_str
    elif dash_count == 1:
        return ver_str.split('-')[0] + "+dirty"
    elif dash_count == 2:
        tag, commits, sha1 = ver_str.split('-')
        return "{}.post0.dev{}+{}".format(tag, commits, sha1)
    elif dash_count == 3:
        tag, commits, sha1, _ = ver_str.split('-')
        return "{}.post0.dev{}+{}.dirty".format(tag, commits, sha1)
    else:
        raise RuntimeError("Invalid version string")


def get_versions(verbose=False):
    vcs_kwds = {"refnames": git_refnames, "full": git_full}
    parentdir = package + '-'
    root = __location__
    # pkg_path is the relative path from the top of the source
    # tree (where the .git directory might live) to this file.
    # Invert this to find the root of our package.
    for _ in pkg_path.split(os.sep):
        root = os.path.dirname(root)

    # different version retrieval methods as (method, args, comment)
    ver_retrieval = [
        (version_from_keywords, (vcs_kwds, tag_prefix, verbose),
         'expanded keywords'),
        (version_from_parentdir, (parentdir, root, verbose), 'parentdir'),
        (version_from_git, (tag_prefix, root, verbose), 'git')
    ]
    for method, args, comment in ver_retrieval:
        ver = method(*args)
        if ver:
            if verbose:
                print("got version from {}".format(comment))
            break
    else:
        ver = {"version": "unknown", "full": ""}
    ver['version'] = git2pep440(ver['version'])
    return ver



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings
try:
    import matplotlib.pyplot as plt
    import matplotlib as mp
except ImportError:
    warnings.warn("Matplotlib is necessary for plotting grids.")
try:
    from mpl_toolkits.basemap import Basemap
except ImportError:
    warnings.warn("Basemap is necessary for plotting grids.")
import numpy as np
import pygeogrids.grids as grids


def plot_cell_grid_partitioning(output,
                                cellsize_lon=5.,
                                cellsize_lat=5.0,
                                figsize=(12, 6)):
    """
    Plot an overview of a global cell partitioning.

    Parameters
    ----------
    output: string
        output file name
    """
    mp.rcParams['font.size'] = 10
    mp.rcParams['text.usetex'] = True
    plt.figure(figsize=figsize, dpi=300)
    ax = plt.axes([0, 0, 1, 1])

    map = Basemap(projection="cyl", llcrnrlat=-90, urcrnrlat=90,
                  llcrnrlon=-180, urcrnrlon=180, ax=ax)
    map.drawparallels(np.arange(-90, 90, cellsize_lat), labels=[1, 0, 0, 0],
                      linewidth=0.5)
    map.drawmeridians(np.arange(-180, 180, cellsize_lon),
                      labels=[0, 0, 0, 1], rotation='vertical', linewidth=0.5)
    # fill continents 'coral' (with zorder=0), color wet areas 'aqua'
    map.drawmapboundary(fill_color='aqua')
    map.fillcontinents(color='0.6', lake_color='aqua')
    label_lats = np.arange(-90 + cellsize_lat / 2., 90, cellsize_lat)
    label_lons = np.arange(-180 + cellsize_lon / 2., 180, cellsize_lon)
    lons, lats = np.meshgrid(label_lons, label_lats)
    x, y = map(lons.flatten(), lats.flatten())
    cells = grids.lonlat2cell(lons.flatten(), lats.flatten(),
                              cellsize_lon=cellsize_lon, cellsize_lat=cellsize_lat)
    for xt, yt, cell in zip(x, y, cells):
        plt.text(xt, yt, "{:}".format(cell), fontsize=4,
                 va="center", ha="center", weight="bold")
    plt.savefig(output, format='png', dpi=300)
    plt.close()



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Module for saving grid to netCDF.
"""

from netCDF4 import Dataset
import numpy as np
import os
from datetime import datetime
from pygeogrids import CellGrid, BasicGrid


def save_lonlat(filename, arrlon, arrlat, geodatum, arrcell=None,
                gpis=None, subsets={}, global_attrs=None,
                format='NETCDF4',
                zlib=False,
                complevel=4,
                shuffle=True):
    """
    saves grid information to netCDF file

    Parameters
    ----------
    filename : string
        name of file
    arrlon : numpy.array
        array of longitudes
    arrlat : numpy.array
        array of latitudes
    geodatum : object
        pygeogrids.geodetic_datum.GeodeticDatum object associated with lon/lat
    arrcell : numpy.array, optional
        array of cell numbers
    gpis : numpy.array, optional
        gpi numbers if not index of arrlon, arrlat
    subsets : dict of dicts, optional
        keys : long_name of the netcdf variables
        values : dict with the following keys: points, meaning
        e.g. subsets = {'subset_flag': {'points': numpy.array,
                                        'value': int,
                                        'meaning': 'water, land'}}
    global_attrs : dict, optional
        if given will be written as global attributs into netCDF file
    format: string, optional
        choose either from one of these NetCDF formats
                        'NETCDF4'
                        'NETCDF4_CLASSIC'
                        'NETCDF3_CLASSIC'
                        'NETCDF3_64BIT_OFFSET'
    zlib: boolean, optional
        see netCDF documentation
    shuffle: boolean, optional
        see netCDF documentation
    complevel: int, opational
        see netCDF documentation
    """

    with Dataset(filename, 'w', format=format) as ncfile:

        if (global_attrs is not None and 'shape' in global_attrs and
                type(global_attrs['shape']) is not int and
                len(global_attrs['shape']) == 2):

            latsize = global_attrs['shape'][0]
            lonsize = global_attrs['shape'][1]
            ncfile.createDimension("lat", latsize)
            ncfile.createDimension("lon", lonsize)
            gpisize = global_attrs['shape'][0] * global_attrs['shape'][1]
            if gpis is None:
                gpivalues = np.arange(gpisize,
                                      dtype=np.int32).reshape(latsize,
                                                              lonsize)
            else:
                gpivalues = gpis.reshape(latsize, lonsize)

            lons = arrlon.reshape(latsize, lonsize)
            lats = arrlat.reshape(latsize, lonsize)
            # sort arrlon, arrlat and gpis
            arrlon_sorted, arrlat_sorted, gpivalues = sort_for_netcdf(
                lons, lats, gpivalues)

            # sorts arrlat descending
            arrlat_store = np.unique(arrlat_sorted)[::-1]
            arrlon_store = np.unique(arrlon_sorted)

        else:
            ncfile.createDimension("gp", arrlon.size)
            gpisize = arrlon.size
            if gpis is None:
                gpivalues = np.arange(arrlon.size, dtype=np.int32)
            else:
                gpivalues = gpis
            arrlon_store = arrlon
            arrlat_store = arrlat

        dim = list(ncfile.dimensions.keys())

        crs = ncfile.createVariable('crs', np.dtype('int32').char,
                                    shuffle=shuffle,
                                    zlib=zlib, complevel=complevel)
        setattr(crs, 'grid_mapping_name', 'latitude_longitude')
        setattr(crs, 'longitude_of_prime_meridian', 0.)
        setattr(crs, 'semi_major_axis', geodatum.geod.a)
        setattr(crs, 'inverse_flattening', 1. / geodatum.geod.f)
        setattr(crs, 'ellipsoid_name', geodatum.name)

        gpi = ncfile.createVariable('gpi', np.dtype('int32').char, dim,
                                    shuffle=shuffle,
                                    zlib=zlib, complevel=complevel)

        if gpis is None:
            gpi[:] = gpivalues
            setattr(gpi, 'long_name', 'Grid point index')
            setattr(gpi, 'units', '')
            setattr(gpi, 'valid_range', [0, gpisize])
            gpidirect = 0x1b
        else:
            gpi[:] = gpivalues
            setattr(gpi, 'long_name', 'Grid point index')
            setattr(gpi, 'units', '')
            setattr(gpi, 'valid_range', [np.min(gpivalues), np.max(gpivalues)])
            gpidirect = 0x0b

        latitude = ncfile.createVariable('lat', np.dtype('float64').char,
                                         dim[0],
                                         shuffle=shuffle,
                                         zlib=zlib, complevel=complevel)
        latitude[:] = arrlat_store
        setattr(latitude, 'long_name', 'Latitude')
        setattr(latitude, 'units', 'degree_north')
        setattr(latitude, 'standard_name', 'latitude')
        setattr(latitude, 'valid_range', [-90.0, 90.0])

        if len(dim) == 2:
            londim = dim[1]
        else:
            londim = dim[0]
        longitude = ncfile.createVariable('lon', np.dtype('float64').char,
                                          londim,
                                          shuffle=shuffle,
                                          zlib=zlib, complevel=complevel)
        longitude[:] = arrlon_store
        setattr(longitude, 'long_name', 'Longitude')
        setattr(longitude, 'units', 'degree_east')
        setattr(longitude, 'standard_name', 'longitude')
        setattr(longitude, 'valid_range', [-180.0, 180.0])

        if arrcell is not None:
            cell = ncfile.createVariable('cell', np.dtype('int16').char,
                                         dim,
                                         shuffle=shuffle,
                                         zlib=zlib, complevel=complevel)

            if len(dim) == 2:
                arrcell = arrcell.reshape(latsize,
                                          lonsize)
                _, _, arrcell = sort_for_netcdf(lons, lats, arrcell)
            cell[:] = arrcell
            setattr(cell, 'long_name', 'Cell')
            setattr(cell, 'units', '')
            setattr(cell, 'valid_range', [np.min(arrcell), np.max(arrcell)])

        if subsets:
            for subset_name in subsets.keys():
                flag = ncfile.createVariable(subset_name, np.dtype('int8').char,
                                             dim,
                                             shuffle=shuffle,
                                             zlib=zlib, complevel=complevel)

                # create flag array based on shape of data
                lf = np.zeros_like(gpivalues)
                if len(dim) == 2:
                    lf = lf.flatten()
                value = subsets[subset_name]['value']
                lf[subsets[subset_name]['points']] = value
                if len(dim) == 2:
                    lf = lf.reshape(latsize, lonsize)
                    _, _, lf = sort_for_netcdf(lons, lats, lf)

                flag[:] = lf
                setattr(flag, 'long_name', subset_name)
                setattr(flag, 'units', '')
                setattr(flag, 'coordinates', 'lat lon')
                setattr(flag, 'flag_values', np.arange(2, dtype=np.int8))
                setattr(flag, 'flag_meanings', subsets[subset_name]['meaning'])
                setattr(flag, 'valid_range', [0, value])

        s = "%Y-%m-%d %H:%M:%S"
        date_created = datetime.now().strftime(s)

        attr = {'Conventions': 'CF-1.6',
                'id': os.path.split(filename)[1],  # file name
                'date_created': date_created,
                'geospatial_lat_min': np.round(np.min(arrlat), 4),
                'geospatial_lat_max': np.round(np.max(arrlat), 4),
                'geospatial_lon_min': np.round(np.min(arrlon), 4),
                'geospatial_lon_max': np.round(np.max(arrlon), 4),
                'gpidirect': gpidirect
                }

        ncfile.setncatts(attr)

        if global_attrs is not None:
            ncfile.setncatts(global_attrs)


def sort_for_netcdf(lons, lats, values):
    """
    Sort an 2D array for storage in a netCDF file.
    This mans that the latitudes are stored from
    90 to -90 and the longitudes from -180 to 180.
    Input arrays have to have shape latdim, londim
    which would mean for a global 10 degree grid (18, 36).

    Parameters
    ----------
    lons: numpy.ndarray
        2D numpy array of longitudes
    lats: numpy.ndarray
        2D numpy array of latitudes
    values: numpy.ndarray
        2D numpy array of values to sort

    Returns
    -------
    lons: numpy.ndarray
        2D numpy array of longitudes, sorted
    lats: numpy.ndarray
        2D numpy array of latitudes, sorted
    values: numpy.ndarray
        2D numpy array of values to sort, sorted
    """

    arrlat = lats.flatten()
    arrlon = lons.flatten()
    arrval = values.flatten()
    idxlatsrt = np.argsort(arrlat)[::-1]
    idxlat = np.argsort(arrlat[idxlatsrt].
                        reshape(lats.shape),
                        axis=0)[::-1]
    idxlon = np.argsort(arrlon[idxlatsrt].
                        reshape(lons.shape)
                        [idxlat, np.arange(lons.shape[1])], axis=1)

    values = arrval[idxlatsrt].reshape(
        *lons.shape)[idxlat, np.arange(lons.shape[1])][np.arange(lons.shape[0])[:, None], idxlon]
    lons = arrlon[idxlatsrt].reshape(
        *lons.shape)[idxlat, np.arange(lons.shape[1])][np.arange(lons.shape[0])[:, None], idxlon]
    lats = arrlat[idxlatsrt].reshape(
        *lons.shape)[idxlat, np.arange(lons.shape[1])][np.arange(lons.shape[0])[:, None], idxlon]
    return lons, lats, values


def save_grid(filename, grid, subset_name='subset_flag', subset_value=1.,
              subset_meaning='water land', global_attrs=None):
    """
    save a BasicGrid or CellGrid to netCDF
    it is assumed that a subset should be used as land_points

    Parameters
    ----------
    filename : string
        name of file
    grid : BasicGrid or CellGrid object
        grid whose definition to save to netCDF
    subset_name : string, optional (default: 'subset_flag')
        long_name of the netcdf variable
        if the subset symbolises something other than a land/sea mask
    subset_value: float, optional (default: 1.)
        Value that that the subgrid is written down as in the file.
    subset_meaning : string, optional (default: 'water land')
        will be written into flag_meanings metadata of variable 'subset_name'
    global_attrs : dict, optional
        if given will be written as global attributes into netCDF file
    """
    try:
        arrcell = grid.arrcell
    except AttributeError:
        arrcell = None

    gpis = grid.gpis

    if grid.shape is not None:
        if global_attrs is None:
            global_attrs = {}
        global_attrs['shape'] = grid.shape

    if grid.subset is not None:
        subsets = {subset_name: {
            'points': grid.subset, 'meaning': subset_meaning, 'value': subset_value}}
    else:
        subsets = None

    save_lonlat(filename, grid.arrlon, grid.arrlat, grid.geodatum,
                arrcell=arrcell, gpis=gpis, subsets=subsets, zlib=True,
                global_attrs=global_attrs)


def load_grid(filename, subset_flag='subset_flag', subset_value=1,
              location_var_name='gpi'):
    """
    load a grid from netCDF file

    Parameters
    ----------
    filename : string
        filename
    subset_flag : string, optional (default: 'subset_flag')
        name of the subset to load.
    subset_value : int or list, optional (default: 1)
        Value(s) of the subset variable that points are loaded for.
    location_var_name: string, optional (default: 'gpi')
        variable name under which the grid point locations
        are stored

    Returns
    -------
    grid : BasicGrid or CellGrid instance
        grid instance initialized with the loaded data
    """

    with Dataset(filename, 'r') as nc_data:
        # determine if it is a cell grid or a basic grid
        arrcell = None
        if 'cell' in nc_data.variables.keys():
            arrcell = nc_data.variables['cell'][:].flatten()

        gpis = nc_data.variables[location_var_name][:].flatten()

        shape = None
        if hasattr(nc_data, 'shape'):
            try:
                shape = tuple(nc_data.shape)
            except TypeError as e:
                try:
                    length = len(nc_data.shape)
                except TypeError:
                    length = nc_data.shape.size
                if length == 1:
                    shape = tuple([nc_data.shape])
                else:
                    raise e

        subset = None
        # some old grid do not have a shape attribute
        # this meant that they had shape of len 1
        if shape is None:
            shape = tuple([len(nc_data.variables['lon'][:])])

        # check if grid has regular shape
        if len(shape) == 2:
            lons, lats = np.meshgrid(nc_data.variables['lon'][:],
                                     nc_data.variables['lat'][:])
            lons = lons.flatten()
            lats = lats.flatten()

            if subset_flag in nc_data.variables.keys():
                subset = np.where(
                    np.isin(nc_data.variables[subset_flag][:].flatten(), subset_value))[0]

        elif len(shape) == 1:
            lons = nc_data.variables['lon'][:]
            lats = nc_data.variables['lat'][:]

            # determine if it has a subset
            if subset_flag in nc_data.variables.keys():
                subset = np.where(
                    np.isin(nc_data.variables[subset_flag][:].flatten(), subset_value))[0]

        if 'crs' in nc_data.variables:
            geodatumName = nc_data.variables['crs'].getncattr('ellipsoid_name')
        else:
            # ellipsoid information is missing, use WGS84 by default
            geodatumName = 'WGS84'

        if arrcell is None:
            # BasicGrid
            return BasicGrid(lons,
                             lats,
                             gpis=gpis,
                             geodatum=geodatumName,
                             subset=subset,
                             shape=shape)
        else:
            # CellGrid
            return CellGrid(lons,
                            lats,
                            arrcell,
                            gpis=gpis,
                            geodatum=geodatumName,
                            subset=subset,
                            shape=shape)



from .grids import BasicGrid, CellGrid, genreg_grid, lonlat2cell, reorder_to_cellsize
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except:
    __version__ = 'unknown'



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Testing grid functionality.
"""

import unittest
import numpy.testing as nptest
import numpy as np
from osgeo import ogr
import pytest
import warnings

from pygeogrids.grids import lonlat2cell, BasicGrid
import pygeogrids as grids


class Test_lonlat2cell(unittest.TestCase):

    def setUp(self):
        lat = np.arange(-90, 90, 2.5)
        lon = np.arange(-180, 180, 2.5)
        self.lons, self.lats = np.meshgrid(lon, lat)

    def testlonlat2cell_hist(self):
        """
        Setup grid with unequal cell size along lat and lon and test if the
        correct number of points lay in each cell.
        """
        cells = lonlat2cell(
            self.lons, self.lats, cellsize_lon=15, cellsize_lat=30)
        hist, bin_edges = np.histogram(
            cells.flatten(), bins=len(np.unique(cells)))
        nptest.assert_allclose(hist, np.zeros_like(hist) + 72)

    def testlonlat2cell_edge(self):
        """
        Use points on the 180 degree longitude and see if they fall into
        the correct cell
        """
        lats = [69.8242, 69.122, 68.42]
        lons = [180, 180, 180]
        cells = lonlat2cell(lons, lats)
        assert list(cells) == [31, 31, 31]


class TestFindNearestNeighbor(unittest.TestCase):

    def setUp(self):
        self.grid = grids.genreg_grid(1, 1)

    def test_nearest_neighbor(self):
        gpi, dist = self.grid.find_nearest_gpi(14.3, 18.5)
        assert gpi == 25754
        assert len([dist]) == 1
        lon, lat = self.grid.gpi2lonlat(gpi)
        assert lon == 14.5
        assert lat == 18.5

    def test_nearest_neighbor_list(self):
        gpi, dist = self.grid.find_nearest_gpi([145.1, 90.2], [45.8, -16.3])
        assert len(gpi) == 2
        assert len(dist) == 2
        assert gpi[0] == 16165
        assert gpi[1] == 38430
        lon, lat = self.grid.gpi2lonlat(gpi)
        assert lon[0] == 145.5
        assert lon[1] == 90.5
        assert lat[0] == 45.5
        assert lat[1] == -16.5

    def test_nearest_neighbor_ndarray(self):
        gpi, dist = self.grid.find_nearest_gpi(
            np.array([145.1, 90.2]), np.array([45.8, -16.3]))
        assert len(gpi) == 2
        assert len(dist) == 2
        assert gpi[0] == 16165
        assert gpi[1] == 38430
        lon, lat = self.grid.gpi2lonlat(gpi)
        assert lon[0] == 145.5
        assert lon[1] == 90.5
        assert lat[0] == 45.5
        assert lat[1] == -16.5

    def test_nearest_neighbor_numpy_single(self):
        gpi, dist = self.grid.find_nearest_gpi(
            np.array([145.1, 90.2])[0], np.array([45.8, -16.3])[0])
        assert gpi == 16165
        lon, lat = self.grid.gpi2lonlat(gpi)
        assert lon == 145.5
        assert lat == 45.5

    def test_k_nearest_neighbor(self):
        gpi, dist = self.grid.find_k_nearest_gpi(14.3, 18.5, k=2)
        assert gpi[0, 0] == 25754
        assert gpi[0, 1] == 25753
        assert dist.shape == (1, 2)
        lon, lat = self.grid.gpi2lonlat(gpi[0, 0])
        assert lon == 14.5
        assert lat == 18.5
        lon, lat = self.grid.gpi2lonlat(gpi[0, 1])
        assert lon == 13.5
        assert lat == 18.5

    def test_k_nearest_neighbor_list(self):
        gpi, dist = self.grid.find_k_nearest_gpi(
            [145.1, 90.2], [45.8, -16.3], k=2)
        assert gpi.shape == (2, 2)
        assert dist.shape == (2, 2)
        assert gpi[0, 0] == 16165
        assert gpi[0, 1] == 16164
        assert gpi[1, 0] == 38430
        assert gpi[1, 1] == 38429

    def test_nearest_neighbor_max_dist(self):
        # test with maxdist higher than nearest point
        gpi, dist = self.grid.find_nearest_gpi(14.3, 18.5, max_dist=100e3)
        assert gpi == 25754
        assert len([dist]) == 1
        lon, lat = self.grid.gpi2lonlat(gpi)
        assert lon == 14.5
        assert lat == 18.5

        # test with maxdist lower than nearest point
        gpi, dist = self.grid.find_nearest_gpi(14.3, 18.5, max_dist=10e3)
        assert len(gpi) == 0
        assert len(dist) == 0

class TestCellGridNotGpiDirect(unittest.TestCase):

    """
    Setup simple 2D grid 2.5 degree global grid (144x72) which starts at the
    North Western corner of 90 -180 Test for cell specific features.
    """

    def setUp(self):
        self.latdim = np.arange(90, -90, -2.5)
        self.londim = np.arange(-180, 180, 2.5)
        self.lon, self.lat = np.meshgrid(self.londim, self.latdim)
        self.grid = grids.BasicGrid(self.lon.flatten(), self.lat.flatten(),
                                    gpis=np.arange(self.lon.flatten().size),
                                    shape=(len(self.latdim),
                                           len(self.londim)))
        self.reverse_gpi_grid = grids.BasicGrid(
            self.lon.flatten(), self.lat.flatten(),
            gpis=np.arange(self.lon.flatten().size)[::-1],
            shape=(len(self.latdim),
                   len(self.londim)))
        self.cellgrid = self.grid.to_cell_grid()

    def test_gpi2cell(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        gpi = 200
        cell = self.cellgrid.gpi2cell(gpi)
        assert cell == 1043

    def test_gpi2cell_iterable(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        gpi = [200, 255]
        cell = self.cellgrid.gpi2cell(gpi)
        assert np.all(cell == [1043, 2015])

    def test_gpi2cell_numpy(self):
        """
        test if gpi to cell lookup works correctly
        """
        gpi = np.array([200, 255])
        cell = self.cellgrid.gpi2cell(gpi)
        assert np.all(cell == [1043, 2015])

    def test_gpi2cell_numpy_single(self):
        """
        test if gpi to row column lookup works correctly
        """
        gpi = np.array([200, 255])[0]
        cell = self.cellgrid.gpi2cell(gpi)
        assert cell == 1043

    def test_calc_lut(self):
        """
        Test calcuation of lookuptable into reverse gpi grid.
        This must result in a lookuptable that reverses the gpis.
        """
        lut = self.grid.calc_lut(self.reverse_gpi_grid)
        nptest.assert_allclose(lut[::-1], self.grid.gpis)

    def test_gpi2cell_custom_gpis(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        self.custom_gpi_grid = \
            grids.BasicGrid(self.lon.flatten(), self.lat.flatten(),
                            shape=(len(self.latdim),
                                   len(self.londim)),
                            gpis=np.arange(len(self.lat.flatten()))[::-1])
        self.custom_gpi_cell_grid = self.custom_gpi_grid.to_cell_grid()
        gpi = [200, 255]
        cell = self.custom_gpi_cell_grid.gpi2cell(gpi)
        assert np.all(cell == [1549, 577])
        gpi = 200
        cell = self.custom_gpi_cell_grid.gpi2cell(gpi)
        assert cell == 1549

    def test_subgrid_from_cells(self):
        """
        Test subgrid selection.
        """
        cells = [1549, 577]
        subgrid = self.cellgrid.subgrid_from_cells(cells)
        assert type(subgrid) == type(self.cellgrid)

        for cell in cells:
            gpis, lons, lats = subgrid.grid_points_for_cell(cell)
            cell_index = np.where(cell == self.cellgrid.activearrcell)
            orig_gpis = self.cellgrid.activegpis[cell_index]
            orig_lons = self.cellgrid.activearrlon[cell_index]
            orig_lats = self.cellgrid.activearrlat[cell_index]
            nptest.assert_array_equal(gpis, orig_gpis)
            nptest.assert_array_equal(lons, orig_lons)
            nptest.assert_array_equal(lats, orig_lats)

    def test_subgrid_from_gpis(self):
        """
        Test subgrid selection.
        """
        gpis = [200, 255]
        subgrid = self.cellgrid.subgrid_from_gpis(gpis)
        assert type(subgrid) == type(self.cellgrid)
        lons_should, lats_should = self.cellgrid.gpi2lonlat(gpis)
        cells_should = self.cellgrid.gpi2cell(gpis)
        subgrid_should = grids.CellGrid(
            lons_should, lats_should, cells_should, gpis=gpis)
        assert subgrid == subgrid_should


class TestLutCalculation(unittest.TestCase):

    def setUp(self):
        """
        Setup two grids with similar gpis but with different subset/gpi ordering.
        The lookup tables should still give the correct results.
        The gpi's of the two grids are identical.
        """
        self.lats = np.array([1, 2, 3, 4])
        self.lons = np.array([1, 2, 3, 4])
        self.gpis = [0, 1, 2, 3]
        self.subset = [3, 2]
        self.lats2 = np.array([3, 4, 2, 1])
        self.lons2 = np.array([3, 4, 2, 1])
        self.gpis2 = [2, 3, 1, 0]
        self.subset2 = [0, 1]
        self.grid1 = grids.BasicGrid(self.lons, self.lats, gpis=self.gpis,
                                     subset=self.subset)
        self.grid2 = grids.BasicGrid(self.lons2, self.lats2, gpis=self.gpis2,
                                     subset=self.subset2)

    def test_calc_lut(self):
        lut = self.grid1.calc_lut(self.grid2)
        nptest.assert_array_equal(lut, [-1, -1, 2, 3])
        nptest.assert_array_equal(
            lut[self.grid2.activegpis], self.grid2.activegpis)
        lut2 = self.grid2.calc_lut(self.grid1)
        nptest.assert_array_equal(lut2, [-1, -1, 2, 3])
        nptest.assert_array_equal(
            lut2[self.grid1.activegpis], self.grid1.activegpis)


class TestCellGridNotGpiDirectSubset(unittest.TestCase):

    """Setup simple 2D grid 2.5 degree global grid (144x72) which starts at the
    North Western corner of 90 -180 Test for cell specific features. This grid
    also has a subset with only the first half of points active.
    """

    def setUp(self):
        self.latdim = np.arange(90, -90, -2.5)
        self.londim = np.arange(-180, 180, 2.5)
        self.lon, self.lat = np.meshgrid(self.londim, self.latdim)
        self.grid = grids.BasicGrid(self.lon.flatten(), self.lat.flatten(),
                                    gpis=np.arange(self.lon.flatten().size),
                                    shape=(len(self.londim),
                                           len(self.latdim)),
                                    subset=np.arange(self.lon.flatten().size / 2,
                                                     dtype=np.int))
        self.cellgrid = self.grid.to_cell_grid()

    def test_gpi2cell(self):
        """
        Test if gpi to cell lookup works correctly.
        """
        gpi = 5185
        cell = self.cellgrid.gpi2cell(gpi)
        assert cell == 18

    def test_gpi2cell_iterable(self):
        """
        Test if gpi to cell lookup works correctly.
        """
        gpi = [200, 5185]
        cell = self.cellgrid.gpi2cell(gpi)
        assert np.all(cell == [1043, 18])

    def test_gpi2cell_numpy_single(self):
        """
        test if gpi to cell lookup works correctly
        """
        gpi = np.array([5185, 255])[0]
        cell = self.cellgrid.gpi2cell(gpi)
        assert cell == 18

    def test_gpi2lonlat(self):
        """
        Test if gpi to lon lat lookup works correctly.
        """
        gpi = 5185
        lon, lat = self.cellgrid.gpi2lonlat(gpi)
        assert lon == -177.5
        assert lat == 0.0

    def test_gpi2lonlat_iterable(self):
        """
        Test if gpi to lon lat lookup works correctly.
        """
        gpi = [200, 5185]
        lon, lat = self.cellgrid.gpi2lonlat(gpi)
        assert np.all(lon == [-40.0, -177.5])
        assert np.all(lat == [87.5, 0.0])

    def test_gpi2lonlat_numpy_single(self):
        """
        test if gpi to lon lat lookup works correctly
        """
        gpi = np.array([5185, 255])[0]
        lon, lat = self.cellgrid.gpi2lonlat(gpi)
        assert lon == -177.5
        assert lat == 0.0


class TestCellGrid(unittest.TestCase):

    """
    Setup simple 2D grid 2.5 degree global grid (144x72) which starts at the
    North Western corner of 90 -180 Test for cell specific features.
    """

    def setUp(self):
        self.latdim = np.arange(90, -90, -2.5)
        self.londim = np.arange(-180, 180, 2.5)
        self.lon, self.lat = np.meshgrid(self.londim, self.latdim)
        self.grid = grids.BasicGrid(self.lon.flatten(), self.lat.flatten(),
                                    shape=(len(self.latdim),
                                           len(self.londim)))
        self.cellgrid = self.grid.to_cell_grid()

    def test_gpi2cell(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        gpi = 200
        cell = self.cellgrid.gpi2cell(gpi)
        assert cell == 1043

    def test_gpi2cell_iterable(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        gpi = [200, 255]
        cell = self.cellgrid.gpi2cell(gpi)
        assert np.all(cell == [1043, 2015])

    def test_gpi2cell_numpy_single(self):
        """
        test if gpi to row column lookup works correctly
        """
        gpi = np.array([200, 255])[0]
        cell = self.cellgrid.gpi2cell(gpi)
        assert cell == 1043

    def test_gpi2cell_custom_gpis(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        self.custom_gpi_grid = \
            grids.BasicGrid(self.lon.flatten(), self.lat.flatten(),
                            shape=(len(self.londim),
                                   len(self.latdim)),
                            gpis=np.arange(len(self.lat.flatten()))[::-1])
        self.custom_gpi_cell_grid = self.custom_gpi_grid.to_cell_grid()
        gpi = [200, 255]
        cell = self.custom_gpi_cell_grid.gpi2cell(gpi)
        assert np.all(cell == [1549, 577])
        gpi = 200
        cell = self.custom_gpi_cell_grid.gpi2cell(gpi)
        assert cell == 1549

    def test_subgrid_from_cells(self):
        """
        Test subgrid selection.
        """
        cells = [1549, 577]
        subgrid = self.cellgrid.subgrid_from_cells(cells)
        assert type(subgrid) == type(self.cellgrid)

        for cell in cells:
            gpis, lons, lats = subgrid.grid_points_for_cell(cell)
            cell_index = np.where(cell == self.cellgrid.activearrcell)
            orig_gpis = self.cellgrid.activegpis[cell_index]
            orig_lons = self.cellgrid.activearrlon[cell_index]
            orig_lats = self.cellgrid.activearrlat[cell_index]
            nptest.assert_array_equal(gpis, orig_gpis)
            nptest.assert_array_equal(lons, orig_lons)
            nptest.assert_array_equal(lats, orig_lats)

    def test_subgrid_from_gpis(self):
        """
        Test subgrid selection.
        """
        gpis = [200, 255]
        subgrid = self.cellgrid.subgrid_from_gpis(gpis)
        assert type(subgrid) == type(self.cellgrid)
        lons_should, lats_should = self.cellgrid.gpi2lonlat(gpis)
        cells_should = self.cellgrid.gpi2cell(gpis)
        subgrid_should = grids.CellGrid(
            lons_should, lats_should, cells_should, gpis=gpis)
        assert subgrid == subgrid_should

    def test_get_bbox_grid_points(self):
        gpis = self.cellgrid.get_bbox_grid_points(latmin=-10,
                                                  latmax=-5,
                                                  lonmin=-10,
                                                  lonmax=-5)
        nptest.assert_allclose(gpis,
                               np.array([5684, 5685, 5828, 5829,
                                         5540, 5541, 5686, 5830, 5542]))
        # gpis should come back sorted by cells
        nptest.assert_allclose(self.cellgrid.gpi2cell(gpis),
                               np.array([1240, 1240, 1240, 1240,
                                         1241, 1241, 1276, 1276, 1277]))
        lats, lons = self.cellgrid.get_bbox_grid_points(latmin=-10,
                                                        latmax=-5,
                                                        lonmin=-10,
                                                        lonmax=-5,
                                                        coords=True)
        lats_should = np.array([-7.5, -7.5, -10., -10.,
                                -5., -5., -7.5, -10., -5.])
        lons_should = np.array([-10.,  -7.5, -10.,  -7.5,
                                -10.,  -7.5,  -5.,  -5.,  -5.])
        nptest.assert_allclose(lats,
                               lats_should)
        nptest.assert_allclose(lons,
                               lons_should)
        gpis, lats, lons = self.cellgrid.get_bbox_grid_points(latmin=-10,
                                                              latmax=-5,
                                                              lonmin=-10,
                                                              lonmax=-5,
                                                              both=True)
        lats_should = np.array([-7.5, -7.5, -10., -10.,
                                -5., -5., -7.5, -10., -5.])
        lons_should = np.array([-10.,  -7.5, -10.,  -7.5,
                                -10.,  -7.5,  -5.,  -5.,  -5.])
        nptest.assert_allclose(lats,
                               lats_should)
        nptest.assert_allclose(lons,
                               lons_should)
        nptest.assert_allclose(gpis,
                               np.array([5684, 5685, 5828, 5829,
                                         5540, 5541, 5686, 5830, 5542]))
        # gpis should come back sorted by cells
        nptest.assert_allclose(self.cellgrid.gpi2cell(gpis),
                               np.array([1240, 1240, 1240, 1240,
                                         1241, 1241, 1276, 1276, 1277]))


def test_setup_grid_with_lists():

    grid = grids.BasicGrid([1, 2, 3, 4, 5], [1, 2, 3, 4, 5])

    nptest.assert_allclose(grid.arrlon, np.array([1, 2, 3, 4, 5]))
    nptest.assert_allclose(grid.arrlat, np.array([1, 2, 3, 4, 5]))


def test_setup_cellgrid_with_lists():

    grid = grids.CellGrid([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [1, 1, 1, 1, 1])

    nptest.assert_allclose(grid.arrlon, np.array([1, 2, 3, 4, 5]))
    nptest.assert_allclose(grid.arrlat, np.array([1, 2, 3, 4, 5]))
    nptest.assert_allclose(grid.arrcell, np.array([1, 1, 1, 1, 1]))


class Test_2Dgrid(unittest.TestCase):

    """
    Setup simple 2D grid 2.5 degree global grid (144x72) which starts at the
    North Western corner of 90 -180 and test 2D lookup.
    """

    def setUp(self):
        self.latdim = np.arange(90, -90, -2.5)
        self.londim = np.arange(-180, 180, 2.5)
        self.lon, self.lat = np.meshgrid(self.londim, self.latdim)
        self.grid = grids.BasicGrid(self.lon.flatten(), self.lat.flatten(),
                                    shape=(len(self.latdim),
                                           len(self.londim)))

    def test_gpi2rowcol(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        gpi = 200
        row_should = 1
        column_should = 200 - 144
        row, column = self.grid.gpi2rowcol(gpi)
        assert row == row_should
        assert column == column_should

    def test_gpi2rowcol_np_int(self):
        """
        test if gpi to row column lookup works correctly
        """
        gpi = np.array([200])[0]
        row_should = 1
        column_should = 200 - 144
        row, column = self.grid.gpi2rowcol(gpi)
        assert row == row_should
        assert column == column_should

    def test_gpi2rowcol_iterable(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        gpi = [143, 200, 255]
        row_should = [0, 1, 1]
        column_should = [143, 200 - 144, 255 - 144]
        row, column = self.grid.gpi2rowcol(gpi)
        assert np.all(row == row_should)
        assert np.all(column == column_should)

    def test_gpi2rowcol_custom_gpis(self):
        """
        Test if gpi to row column lookup works correctly.
        """
        self.custom_gpi_grid = grids.BasicGrid(self.lon.flatten(),
                                               self.lat.flatten(),
                                               shape=(len(self.latdim),
                                                      len(self.londim)),
                                               gpis=np.arange(len(self.lat.flatten()))[::-1])
        gpi = [200, 255]
        row_should = [70, 70]
        column_should = [87, 32]
        row, column = self.custom_gpi_grid.gpi2rowcol(gpi)
        assert np.all(row == row_should)
        assert np.all(column == column_should)

    def test_gpi2lonlat(self):
        """
        Test if gpi to longitude latitude lookup works correctly.
        """
        gpi = 200
        lat_should = 87.5
        lon_should = -180 + (200 - 144) * 2.5
        lon, lat = self.grid.gpi2lonlat(gpi)
        assert lon == lon_should
        assert lat == lat_should

    def test_lonlat2d(self):
        """
        Test if lonlat 2d grids are the same as the grids used for making the grid.
        """
        assert np.all(self.lon == self.grid.lon2d)
        assert np.all(self.lat == self.grid.lat2d)

    def test_tocellgrid(self):
        """
        test if to_cell_grid method works correctly
        """
        cell_grid = self.grid.to_cell_grid()
        result = grids.BasicGrid.__eq__(self.grid, cell_grid)
        assert result


def test_genreggrid():
    """
    Test generation of regular grids.
    """
    grid = grids.genreg_grid()
    assert grid.shape == (180, 360)
    lon, lat = grid.gpi2lonlat(3)
    assert lon == -176.5
    assert lat == 89.5
    lon, lat = grid.gpi2lonlat(360)
    assert lon == -179.5
    assert lat == 88.5


def test_reorder_to_cellsize():
    """
    Test reordering to different cellsize
    """
    lons = np.array([-177, -177, -176, -176])
    lats = np.array([51, 57, 51, 57])
    gpis = np.array([1, 2, 3, 4])
    cells = np.array([14, 14, 14, 14])
    orig_grid = grids.CellGrid(lons, lats, cells, gpis=gpis)
    reordered_grid = grids.reorder_to_cellsize(orig_grid, 5.0, 5.0)
    nptest.assert_almost_equal(reordered_grid.gpis,
                               np.array([1, 3, 2, 4]))
    nptest.assert_almost_equal(reordered_grid.arrlon,
                               np.array([-177, -176, -177, -176]))
    nptest.assert_almost_equal(reordered_grid.arrlat,
                               np.array([51, 51, 57, 57]))
    nptest.assert_almost_equal(reordered_grid.arrcell,
                               np.array([14, 14, 14, 14]))


class Test_ShpGrid(unittest.TestCase):

    def setUp(self):
        """
        Setup grid and shp_file to check if grid points fall in shp.
        """
        lat = np.arange(-90, 90, 1)
        lon = np.arange(-180, 180, 1)
        self.lons, self.lats = np.meshgrid(lon, lat)
        self.grid = grids.BasicGrid(self.lons.flatten(), self.lats.flatten())

        ring = ogr.Geometry(ogr.wkbLinearRing)
        ring.AddPoint(14, 45)
        ring.AddPoint(14, 47)
        ring.AddPoint(16, 47)
        ring.AddPoint(16, 45)
        ring.AddPoint(14, 45)

        poly = ogr.Geometry(ogr.wkbPolygon)
        poly.AddGeometry(ring)

        self.shp = poly

    def test_shpgrid(self):
        '''
        Check if gridpoints fall in polygon.
        '''
        subgrid = self.grid.get_shp_grid_points(self.shp)
        assert subgrid.activearrlon == 15
        assert subgrid.activearrlat == 46


@pytest.mark.filterwarnings("error")
def test_BasicGrid_transform_lon():
    """
    Tests whether transforming longitudes works as expected.
    """

    lat = np.asarray([10, -10, 5, 42])
    lon_pos = np.asarray([0, 90, 180, 270])
    lon_centered = np.asarray([0, 90, 180, -90])

    # case 1: warning and transformation
    with pytest.warns(UserWarning):
        grid = BasicGrid(lon_pos, lat)
        assert np.all(grid.arrlon == lon_centered)

    # case 2: no warning and transform
    grid = BasicGrid(lon_pos, lat, transform_lon=True)
    assert np.all(grid.arrlon == lon_centered)

    # case 3: no warning and no transform
    grid = BasicGrid(lon_pos, lat, transform_lon=False)
    assert np.all(grid.arrlon == lon_pos)



if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Module for extracting grid points from global administrative areas
"""

try:
    from osgeo import ogr
    ogr_installed = True
except ImportError:
    ogr_installed = False


def get_gad_grid_points(grid, gadm_shp_path, level, name=None, oid=None):
    """
    Returns all grid points located in a administrative area. For this
    function the files from
    http://biogeo.ucdavis.edu/data/gadm2.8/gadm28_levels.shp.zip
    need to be available in the folder gadm_shp_path
    Optinal as coordinates. Currently only works in WGS84.

    Parameters
    ----------
    grid: object
    gadm_shp_path: path
        Location to GADM28 shapefiles
    level: int
        Global Administrative Database Level
        0 : country
        1 : province/county/state/region/municipality/...
        2 : municipality/District/county/...
    name: str
        name of region at indicated level. For countries the english name
    oid: int
        OBJECTID of feature. This only works with the correct level shp.

    Returns
    -------
    grid : BasicGrid
            Subgrid.

    Raises
    ------
    ValueError: If name or oid are not found in shapefile of given level
    ImportError: If gdal or osgeo are not installed
    """
    if ogr_installed:
        drv = ogr.GetDriverByName('ESRI Shapefile')
        ds_in = drv.Open(gadm_shp_path + 'gadm28_adm{:}.shp'.format(level))
        lyr_in = ds_in.GetLayer(0)
        if name:
            if level == 0:
                lyr_in.SetAttributeFilter("NAME_ENGLI = '%s'" % (name))
            else:
                lyr_in.SetAttributeFilter("NAME_%s = '%s'" % (level, name))

        if oid:
            lyr_in.SetAttributeFilter("OBJECTID = '%s'" % (oid))
        if lyr_in.GetFeatureCount() > 0:
            feature = lyr_in.GetNextFeature()
            ply = feature.GetGeometryRef()
            return grid.get_shp_grid_points(ply)
        else:
            raise ValueError("Requested Object not found in shapefile.")

    else:
        raise ImportError("No supported implementation installed."
                          "Please install gdal and osgeo.")



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
The grids module defines the grid classes.
"""

import numpy as np
import numpy.testing as nptest
import warnings
try:
    from osgeo import ogr
    ogr_installed = True
except ImportError:
    ogr_installed = False

try:
    from itertools import izip as zip
except ImportError:
    pass  # python3

import pygeogrids.nearest_neighbor as NN
from pygeogrids.geodetic_datum import GeodeticDatum


class GridDefinitionError(Exception):
    pass


class GridIterationError(Exception):
    pass


class BasicGrid(object):

    """
    Grid that just has lat,lon coordinates and can find the
    nearest neighbour. It can also yield the gpi, lat, lon
    information in order.

    Parameters
    ----------
    lon : numpy.array
        longitudes of the points in the grid
    lat : numpy.array
        latitudes of the points in the grid
    geodatum : basestring
        Name of the geodatic datum associated with the grid
    gpis : numpy.array, optional
        if the gpi numbers are in a different order than the
        lon and lat arrays an array containing the gpi numbers
        can be given
        if no array is given here the lon lat arrays are given
        gpi numbers starting at 0
    subset : numpy.array, optional
        if the active part of the array is only a subset of
        all the points then the subset array which is a index
        into lon and lat can be given here.
    setup_kdTree : boolean, optional
        if set (default) then the kdTree for nearest neighbour
        search will be built on initialization
    shape : tuple, optional
        The shape of the grid array in 2-d space.
        e.g. for a 1x1 degree global regular grid the shape would be (180,360).
        if given the grid can be reshaped into the given shape
        this indicates that it is a regular grid and fills the
        attributes self.lon2d and self.lat2d which
        define the grid only be the meridian coordinates(self.lon2d) and
        the coordinates of the circles of latitude(self.lat2d).
        The shape has to be given as (lat2d, lon2d)
        It it is not given the shape is set to the length of the input
        lon and lat arrays.
    transform_lon : bool or None, optional (default: None)
        Whether to transform longitudes to values between -180 and 180.
        By default values are transformed, but a warning is issued.
        To turn off the warning, set this to ``True``, to turn of
        transformation set this to ``False``.

    Attributes
    ----------
    arrlon : numpy.array
        1D array of all longitudes of the grid
    arrlat : numpy.array
        1D array of all latitudes of the grid
    n_gpi : int
        number of gpis in the grid
    gpidirect : boolean
        if true the gpi number is equal to the index
        of arrlon and arrlat
    gpis : numpy.array
        gpi number for elements in arrlon and arrlat
        gpi[i] is located at arrlon[i],arrlat[i]
    subset : numpy.array
        if given then this contains the indices of a subset of
        the grid. This can be used if only a part of a grid is
        interesting for a application. e.g. land points, or only
        a specific country
    allpoints : boolean
        if False only a subset of the grid is active
    activearrlon : numpy.array
        array of longitudes that are active, is defined by
        arrlon[subset] if a subset is given otherwise equal to
        arrlon
    activearrlat : numpy.array
        array of latitudes that are active, is defined by
        arrlat[subset] if a subset is given otherwise equal to
        arrlat
    activegpis : numpy.array
        array of gpis that are active, is defined by
        gpis[subset] if a subset is given otherwise equal to
        gpis
    geodatum : object
        pygeogrids.geodatic_datum object (reference ellipsoid) associated
        with the grid
    issplit : boolean
        if True then the array was split in n parts with
        the self.split function
    kdTree : object
        grid.nearest_neighbor.findGeoNN object for
        nearest neighbor search
    shape : tuple, optional
        if given during initialization then this is
        the shape the grid can be reshaped to
    lat2d : numpy.array, optional
        if shape is given this attribute contains all latitudes according to
        the provided 2d-shape that make up the grid
    lon2d : numpy.array, optional
        if shape is given this attribute contains all longitudes according to
        the provided 2d-shape that make up the grid
    """

    def __init__(self, lon, lat, gpis=None, geodatum='WGS84', subset=None,
                 setup_kdTree=True, shape=None, transform_lon=None):
        """
        init method, prepares lon and lat arrays for _transform_lonlats if
        necessary

        """
        lon = np.asanyarray(lon)
        lat = np.asanyarray(lat)
        if gpis is not None:
            gpis = np.asanyarray(gpis)
        if subset is not None:
            subset = np.asanyarray(subset)

        if lat.shape != lon.shape:
            raise GridDefinitionError(
                "lat and lon np.arrays have to have equal shapes")

        self.n_gpi = len(lon)

        # transfrom longitudes to be between -180 and 180 if they are between 0
        # and 360
        if transform_lon or transform_lon is None:
            if np.any(lon > 180):
                lon[lon > 180] -= 360
                if transform_lon is None:
                    warnings.warn(
                        "Longitude values have been transformed to be in"
                        " (-180, 180]. If this was not intended or to suppress"
                        " this warning set the transform_lon keyword argument"
                    )

        self.arrlon = lon
        self.arrlat = lat

        self.shape = None

        if shape is not None and len(shape) == 2:
            if len(self.arrlat) % shape[0] != 0:
                msg = ("Given shape does not have the correct first dimension."
                       " Length of lat array is not divisible by shape[0] "
                       " without rest")
                raise GridDefinitionError(msg)

            if len(self.arrlon) % shape[1] != 0:
                msg = ("Given shape does not have the correct second "
                       "dimension. Length of lon array is not divisible by "
                       "shape[1] without rest")
                raise GridDefinitionError(msg)

            self.shape = shape
            self.lat2d = np.reshape(self.arrlat, self.shape)
            self.lon2d = np.reshape(self.arrlon, self.shape)

        else:
            self.shape = tuple([len(self.arrlon)])

        self.geodatum = GeodeticDatum(geodatum)

        if gpis is None:
            self.gpis = np.arange(self.n_gpi, dtype=int)
            self.gpidirect = True
        else:
            if lat.shape != gpis.shape:
                raise GridDefinitionError("lat, lon gpi np.arrays have to "
                                          "have equal shapes")
            self.gpis = gpis
            self.gpidirect = False

        self.subset = subset

        if subset is not None:
            self.activearrlon = self.arrlon[subset]
            self.activearrlat = self.arrlat[subset]
            self.activegpis = self.gpis[subset]
            self.allpoints = False
        else:
            self.activearrlon = self.arrlon
            self.activearrlat = self.arrlat
            self.activegpis = self.gpis
            self.allpoints = True

        self.issplit = False

        self.kdTree = None
        if setup_kdTree:
            self._setup_kdtree()

    def _setup_kdtree(self):
        """
        Setup kdTree
        """
        if self.kdTree is None:
            self.kdTree = NN.findGeoNN(self.activearrlon, self.activearrlat,
                                       self.geodatum)
            self.kdTree._build_kdtree()

    def split(self, n):
        """
        Function splits the grid into n parts this changes not function but
        grid_points() which takes the argument n and will only iterate through
        this part of the grid.

        Parameters
        ----------
        n : int
            Number of parts the grid should be split into
        """
        self.issplit = True
        self.subarrlats = np.array_split(self.activearrlat, n)
        self.subarrlons = np.array_split(self.activearrlon, n)
        self.subgpis = np.array_split(self.activegpis, n)

    def unite(self):
        """
        Unites a split array, so that it can be iterated over as a whole
        again.
        """
        self.issplit = False

    def grid_points(self, *args):
        """
        Yields all grid points in order

        Parameters
        ----------
        n : int, optional
            if the grid is split in n parts using the split function
            then this iterator will only iterate of the nth part of the
            grid

        Returns
        -------
        gpi : long
            grid point index
        lon : float
            longitude of gpi
        lat : float
            longitude of gpi
        """

        if not self.issplit and len(args) == 0:
            return self._normal_grid_points()
        elif self.issplit and len(args) == 1:
            return self._split_grid_points(args[0])

        raise GridIterationError("this function only takes an argument if "
                                 "the grid is split, and takes no argument "
                                 "if the grid is not split")

    def get_grid_points(self, *args):
        """
        Returns all active grid points.

        Parameters
        ----------
        n : int, optional
            if the grid is split in n parts using the split function
            then this function will only return the nth part of the
            grid

        Returns
        -------
        gpis : numpy.ndarray
            Grid point indices.
        arrlon : numpy.ndarray
            Longitudes.
        arrlat : numpy.ndarray
            Latitudes.
        """

        if not self.issplit and len(args) == 0:
            return (self.activegpis,
                    self.activearrlon,
                    self.activearrlat)

        elif self.issplit and len(args) == 1:
            n = args[0]
            return (self.subgpis[n],
                    self.subarrlons[n],
                    self.subarrlats[n])

    def _normal_grid_points(self):
        """
        Yields all grid points in order.

        Returns
        -------
        gpi : long
            grid point index
        lon : float
            longitude of gpi
        lat : float
            longitude of gpi
        """
        for i, (lon, lat) in enumerate(zip(self.activearrlon,
                                           self.activearrlat)):
            yield self.activegpis[i], lon, lat

    def _split_grid_points(self, n):
        """
        Yields all grid points or split grid in order.

        Parameters
        ----------
        n : int
            Number of subgrid to yield

        Returns
        -------
        gpi : long
            Grid point index.
        lon : float
            Longitude of gpi.
        lat : float
            Longitude of gpi.
        """

        for i, (lon, lat) in enumerate(zip(self.subarrlons[n],
                                           self.subarrlats[n])):
            yield self.subgpis[n][i], lon, lat

    def find_nearest_gpi(self, lon, lat, max_dist=np.Inf):
        """
        Finds nearest gpi, builds kdTree if it does not yet exist.

        Parameters
        ----------
        lon : float or iterable
            Longitude of point.
        lat : float or iterable
            Latitude of point.
        max_dist : float, optional
            Maximum distance [m] to consider for search (default: np.Inf).

        Returns
        -------
        gpi : long
            Grid point index.
        distance : float
            Distance of gpi to given lon, lat.
            At the moment not on a great circle but in spherical
            cartesian coordinates.
        """
        gpi, distance = self.find_k_nearest_gpi(lon, lat, max_dist=max_dist, k=1)

        if not _element_iterable(lon) and len(gpi) > 0:
            gpi = gpi[0]
            distance = distance[0]

        return gpi, distance

    def find_k_nearest_gpi(self, lon, lat, max_dist=np.Inf, k=1):
        """
        Find k nearest gpi, builds kdTree if it does not yet exist.

        Parameters
        ----------
        lon : float or iterable
            Longitude of point.
        lat : float or iterable
            Latitude of point.
        max_dist : float, optional
            Maximum distance to consider for search (default: np.Inf).
        k : int, optional
            The number of nearest neighbors to return (default: 1).

        Returns
        -------
        gpi : long
            Grid point index.
        distance : float
            Distance of gpi to given lon, lat.
            At the moment not on a great circle but in spherical
            cartesian coordinates.
        """
        if self.kdTree is None:
            self._setup_kdtree()

        distance, ind = self.kdTree.find_nearest_index(
            lon, lat, max_dist=max_dist, k=k)

        if self.gpidirect and self.allpoints:
            gpi = ind
        else:
            gpi = self.activegpis[ind]

        return gpi, distance

    def gpi2lonlat(self, gpi):
        """
        Longitude and latitude for given gpi.

        Parameters
        ----------
        gpi : int32 or iterable
            Grid point index.

        Returns
        -------
        lon : float
            Longitude of gpi.
        lat : float
            Latitude of gpi
        """
        # check if iterable
        iterable = _element_iterable(gpi)

        gpi = np.atleast_1d(gpi)
        if self.gpidirect:
            lons, lats = self.arrlon[gpi], self.arrlat[gpi]
        else:
            # get the indices that would sort the gpis
            gpisorted = np.argsort(self.gpis)
            # find the position where the gpis fit in the sorted array
            pos = np.searchsorted(self.gpis[gpisorted], gpi)
            index = gpisorted[pos]
            lons, lats = self.arrlon[index], self.arrlat[index]

        if not iterable:
            lons = lons[0]
            lats = lats[0]

        return lons, lats

    def gpi2rowcol(self, gpi):
        """
        If the grid can be reshaped into a sensible 2D shape then this
        function gives the row(latitude dimension) and
        column(longitude dimension) indices of the gpi in the 2D grid.

        Parameters
        ----------
        gpi : int32
            Grid point index.

        Returns
        -------
        row : int
            Row in 2D array.
        col : int
            Column in 2D array.
        """
        # check if iterable
        iterable = _element_iterable(gpi)

        gpi = np.atleast_1d(gpi)
        if len(self.shape) == 2:
            if self.gpidirect:
                index = gpi
            else:
                # get the indices that would sort the gpis
                gpisorted = np.argsort(self.gpis)
                # find the position where the gpis fit in the sorted array
                pos = np.searchsorted(self.gpis[gpisorted], gpi)
                index = gpisorted[pos]

            index_lat = (index / self.shape[1]).astype(np.int)
            index_lon = index % self.shape[1]
            if not iterable:
                index_lat = index_lat[0]
                index_lon = index_lon[0]
            return index_lat, index_lon

        else:
            raise(GridDefinitionError("Grid has no 2D shape"))

    def calc_lut(self, other, max_dist=np.Inf, into_subset=False):
        """
        Takes other BasicGrid or CellGrid objects and computes a lookup table
        between them. The lut will have the size of self.n_gpis and will
        for every grid point have the nearest index into other.arrlon etc.

        Parameters
        ----------
        other : grid object
            to which to calculate the lut to
        max_dist : float, optional
            maximum allowed distance in meters
        into_subset : boolean, optional
            if set the returned lut will have the index into the subset
            if the other grid is a subset of a grid.
            Example:
            if e.g. ind_l is used for the warp_grid some datasets will
            be given as arrays with len(ind_l) elements. These
            datasets can not be indexed with gpi numbers but have to
            be indexed with indices into the subset
        """

        if self.kdTree is None:
            self._setup_kdtree()

        if other.kdTree is None:
            other._setup_kdtree()

        if self.kdTree.kdtree is not None and other.kdTree.kdtree is not None:
            dist, index = other.kdTree.find_nearest_index(
                self.activearrlon, self.activearrlat, max_dist=max_dist)

            valid_index = np.where(dist != np.inf)[0]
            dist = dist[valid_index]
            index = index[valid_index]
            if not other.gpidirect or not other.allpoints:
                if not into_subset:
                    index = other.activegpis[index]

            active_lut = np.empty_like(self.activearrlat, dtype=np.int64)
            active_lut.fill(-1)
            active_lut[valid_index] = index

            if not self.allpoints:
                gpi_lut = np.empty_like(self.gpis)
                gpi_lut.fill(-1)
                gpi_lut[self.gpis[self.subset]] = active_lut
            elif not self.gpidirect:
                gpi_lut = np.empty(np.max(self.activegpis) + 1, dtype=np.int64)
                gpi_lut.fill(-1)
                gpi_lut[self.activegpis] = active_lut
            else:
                gpi_lut = active_lut

            return gpi_lut

    def get_shp_grid_points(self, ply):
        """
        Returns all grid points located in a submitted shapefile,
        optinal as coordinates. Currently only works in WGS84.

        Parameters
        ----------
        ply: object, OGRGeometryShadow
            the Geometry of the Feature as returned from ogr.GetGeometryRef

        Returns
        -------
        grid : BasicGrid
            Subgrid.
        """

        if ogr_installed:
            lonmin, lonmax, latmin, latmax = ply.GetEnvelope()
            gpis, lats, lons = self.get_bbox_grid_points(latmin, latmax,
                                                         lonmin, lonmax,
                                                         both=True)

            lon_ip = []
            lat_ip = []
            gpi_ip = []

            if len(gpis) > 0:
                for gpi, lon, lat in zip(gpis, lons, lats):
                    pt = ogr.Geometry(ogr.wkbPoint)
                    pt.SetPoint_2D(0, float(lon), float(lat))
                    if ply.Contains(pt):
                        lon_ip.append(lon)
                        lat_ip.append(lat)
                        gpi_ip.append(gpi)
            if len(gpi_ip) > 0:
                return self.subgrid_from_gpis(gpi_ip)
            else:
                return None

        else:
            raise Exception("No supported implementation installed.\
                            Please install gdal and osgeo.")

    def get_bbox_grid_points(self, latmin=-90, latmax=90, lonmin=-180,
                             lonmax=180, coords=False, both=False):
        """
        Returns all grid points located in a submitted geographic box,
        optinal as coordinates

        Parameters
        ----------
        latmin : float, optional
            minimum latitude
        latmax : float, optional
            maximum latitude
        lonmin : float, optional
            minimum latitude
        lonmax : float, optional
            maximum latitude
        coords : boolean, optional
            set to True if coordinates should be returned
        both: boolean, optional
            set to True if gpis and coordinates should be returned

        Returns
        -------
        gpi : numpy.ndarray
            grid point indices, if coords=False
        lat : numpy.ndarray
            longitudes of gpis, if coords=True
        lon : numpy.ndarray
            longitudes of gpis, if coords=True
        """

        gp_info = np.array(list(self.grid_points()))
        gpis = gp_info[:, 0].astype(int)
        lons = gp_info[:, 1]
        lats = gp_info[:, 2]
        index = np.where((lats <= latmax) &
                         (lats >= latmin) &
                         (lons <= lonmax) &
                         (lons >= lonmin))

        if coords is True:
            return lats[index], lons[index]
        elif both:
            return gpis[index], lats[index], lons[index]
        else:
            return gpis[index]

    def to_cell_grid(self, cellsize=5.0, cellsize_lat=None, cellsize_lon=None):
        """
        Convert grid to cellgrid with a cell partition of cellsize.

        Parameters
        ----------
        cellsize : float, optional
            Cell size in degrees
        cellsize_lon : float, optional
            Cell size in degrees on the longitude axis
        cellsize_lat : float, optional
            Cell size in degrees on the latitude axis

        Returns
        -------
        cell_grid : CellGrid object
            Cell grid object.
        """
        cells = lonlat2cell(self.arrlon, self.arrlat, cellsize=cellsize,
                            cellsize_lat=cellsize_lat,
                            cellsize_lon=cellsize_lon)

        if self.gpidirect:
            gpis = None
        else:
            gpis = self.gpis

        return CellGrid(self.arrlon, self.arrlat, cells, gpis=gpis,
                        subset=self.subset, shape=self.shape)

    def subgrid_from_gpis(self, gpis):
        """
        Generate a subgrid for given gpis.

        Parameters
        ----------
        gpis : int, numpy.ndarray
            Grid point indices.

        Returns
        -------
        grid : BasicGrid
            Subgrid.
        """
        sublons, sublats = self.gpi2lonlat(gpis)

        return BasicGrid(sublons, sublats, gpis, geodatum=self.geodatum.name)

    def __eq__(self, other):
        """
        Compare arrlon, arrlat, gpis, subsets and shape.

        Returns
        -------
        result : boolean
            Returns True if grids are equal.
        """
        # only test to certain significance for float variables
        # grids are assumed to be the same if the gpi, lon, lat tuples are the
        # same
        idx_gpi = np.argsort(self.gpis)
        idx_gpi_other = np.argsort(other.gpis)
        gpisame = np.all(self.gpis[idx_gpi] == other.gpis[idx_gpi_other])
        try:
            nptest.assert_allclose(self.arrlon[idx_gpi],
                                   other.arrlon[idx_gpi_other])
            lonsame = True
        except AssertionError:
            lonsame = False
        try:
            nptest.assert_allclose(self.arrlat[idx_gpi],
                                   other.arrlat[idx_gpi_other])
            latsame = True
        except AssertionError:
            latsame = False
        if self.subset is not None and other.subset is not None:
            subsetsame = np.all(
                sorted(self.gpis[self.subset]) == sorted(other.gpis[other.subset]))
        elif self.subset is None and other.subset is None:
            subsetsame = True
        else:
            subsetsame = False

        if self.shape is None and other.shape is None:
            shapesame = True
        elif self.shape is not None and other.shape is not None:
            shapesame = self.shape == other.shape
        else:
            shapesame = False

        if self.geodatum.name == other.geodatum.name:
            geosame = True
        else:
            geosame = False

        return np.all([lonsame, latsame, gpisame, subsetsame, shapesame,
                       geosame])


class CellGrid(BasicGrid):

    """
    Grid that has lat,lon coordinates as well as cell informatin. It can find
    nearest neighbour. It can also yield the gpi, lat, lon, cell information
    in cell order. This is important if the data on the grid is saved in cell
    files on disk as we can go through all grid points with optimized
    IO performance.

    Parameters
    ----------
    lon : numpy.ndarray
        Longitudes of the points in the grid.
    lat : numpy.ndarray
        Latitudes of the points in the grid.
    cells : numpy.ndarray
        Of same shape as lon and lat, containing the cell number of each gpi.
    gpis : numpy.ndarray, optional
        If the gpi numbers are in a different order than the lon and lat
        arrays an array containing the gpi numbers can be given.
    subset : numpy.array, optional
        If the active part of the array is only a subset of all the points
        then the subset array which is a index into lon, lat and cells can
        be given here.

    Attributes
    ----------
    arrcell : numpy.ndarray
        Array of cell number with same shape as arrlon, arrlat.
    activearrcell : numpy.ndarray
        Array of longitudes that are active, is defined by arrlon[subset]
        if a subset is given otherwise equal to arrlon.
    """

    def __init__(self, lon, lat, cells, gpis=None, geodatum='WGS84',
                 subset=None, setup_kdTree=False, **kwargs):

        super(CellGrid, self).__init__(lon, lat, gpis=gpis,
                                       geodatum=geodatum, subset=subset,
                                       setup_kdTree=setup_kdTree, **kwargs)

        self.gpi_lut = None
        cells = np.asanyarray(cells)

        if self.arrlon.shape != cells.shape:
            raise GridDefinitionError(
                "lat, lon and cells np.arrays have to have equal shapes")
        self.arrcell = cells

        if subset is not None:
            self.activearrcell = self.arrcell[subset]
        else:
            self.activearrcell = self.arrcell

    def gpi2cell(self, gpi):
        """
        Cell for given gpi.

        Parameters
        ----------
        gpi : int32 or iterable
            Grid point index.

        Returns
        -------
        cell : int or iterable
            Cell number of gpi.
        """
        # check if iterable
        iterable = _element_iterable(gpi)

        gpi = np.atleast_1d(gpi)
        if self.gpidirect:
            cell = self.arrcell[gpi]
        else:
            if self.gpi_lut is None:
                self.gpi_lut = np.zeros(self.gpis.max() + 1, dtype=np.int32)
                self.gpi_lut[self.gpis] = np.arange(self.gpis.size)

            cell = self.arrcell[self.gpi_lut[gpi]]

        if not iterable:
            cell = cell[0]

        return cell

    def get_cells(self):
        """
        Function to get all cell numbers of the grid.

        Returns
        -------
        cells : numpy.ndarray
            Unique cell numbers.
        """
        return np.unique(self.activearrcell)

    def get_grid_points(self, *args):
        """
        Returns all active grid points.

        Parameters
        ----------
        n : int, optional
            If the grid is split in n parts using the split function then this
            function will only return the nth part of the grid.

        Returns
        -------
        gpis : numpy.ndarray
            Grid point indices.
        arrlon : numpy.ndarray
            Longitudes.
        arrlat :numpy.ndarray
            Latitudes.
        cells : numpy.ndarray
            Cell numbers.
        """
        if not self.issplit and len(args) == 0:
            return (self.activegpis,
                    self.activearrlon,
                    self.activearrlat,
                    self.activearrcell)

        elif self.issplit and len(args) == 1:
            n = args[0]
            return (self.subgpis[n],
                    self.subarrlons[n],
                    self.subarrlats[n],
                    self.subcells[n])

    def grid_points_for_cell(self, cells):
        """
        Get all grid points for a given cell number.

        Parameters
        ----------
        cell : int, numpy.ndarray
            Cell numbers.

        Returns
        -------
        gpis : numpy.ndarray
            Gpis belonging to cell.
        lons : numpy.array
            Longitudes belonging to the gpis.
        lats : numpy.array
            Latitudes belonging to the gpis.
        """
        cells = np.atleast_1d(cells)

        gpis = []
        lons = []
        lats = []
        for cell in cells:
            cell_index = np.where(cell == self.activearrcell)
            gpis.append(self.activegpis[cell_index])
            lons.append(self.activearrlon[cell_index])
            lats.append(self.activearrlat[cell_index])

        gpis = np.hstack(gpis)
        lons = np.hstack(lons)
        lats = np.hstack(lats)

        return gpis, lons, lats

    def split(self, n):
        """
        Function splits the grid into n parts this changes not function but
        grid_points() which takes the argument n and will only iterate through
        this part of the grid.

        Parameters
        ----------
        n : int
            Number of parts the grid should be split into.
        """
        self.issplit = True
        # sort by cell number to split correctly
        sorted_index = np.argsort(self.activearrcell)
        self.subarrlats = np.array_split(self.activearrlat[sorted_index], n)
        self.subarrlons = np.array_split(self.activearrlon[sorted_index], n)
        self.subgpis = np.array_split(self.activegpis[sorted_index], n)
        self.subcells = np.array_split(self.activearrcell[sorted_index], n)

    def _normal_grid_points(self):
        """
        Yields all grid points in cell order.

        Returns
        -------
        gpi : long
            Grid point index.
        lon : float
            Longitude of gpi.
        lat : float
            Longitude of gpi.
        cell : int
            Cell number.
        """
        uniq_cells = np.unique(self.activearrcell)

        for cell in uniq_cells:
            cell_gpis = np.where(cell == self.activearrcell)[0]
            for gpi in cell_gpis:
                yield self.activegpis[gpi], self.activearrlon[gpi], \
                    self.activearrlat[gpi], cell

    def _split_grid_points(self, n):
        """
        Yields all grid points in cell order.

        Parameters
        ----------
        n : int
            Number of subgrid to yield.

        Returns
        -------
        gpi : long
            Grid point index.
        lon : float
            Longitude of gpi.
        lat : float
            Latitude of gpi.
        cell : int
            Cell number.
        """
        uniq_cells = np.unique(self.subcells[n])

        for cell in uniq_cells:
            cell_gpis = np.where(cell == self.subcells[n])[0]
            for gpi in cell_gpis:
                yield self.subgpis[n][gpi], self.subarrlons[n][gpi], \
                    self.subarrlats[n][gpi], cell

    def subgrid_from_gpis(self, gpis):
        """
        Generate a subgrid for given gpis.

        Parameters
        ----------
        gpis : int, numpy.ndarray
            Grid point indices.

        Returns
        -------
        grid : BasicGrid
            Subgrid.
        """
        sublons, sublats = self.gpi2lonlat(gpis)
        subcells = self.gpi2cell(gpis)

        return CellGrid(sublons, sublats, subcells, gpis,
                        geodatum=self.geodatum.name)

    def subgrid_from_cells(self, cells):
        """
        Generate a subgrid for given cells.

        Parameters
        ----------
        cells : int, numpy.ndarray
            Cell numbers.

        Returns
        -------
        grid : CellGrid
            Subgrid.
        """
        subgpis, sublons, sublats = self.grid_points_for_cell(cells)
        subcells = self.gpi2cell(subgpis)

        return CellGrid(sublons, sublats, subcells, subgpis,
                        geodatum=self.geodatum.name)

    def __eq__(self, other):
        """
        Compare cells.

        Returns
        -------
        result : boolean
            Returns true if equal.
        """
        basicsame = super(CellGrid, self).__eq__(other)
        idx_gpi = np.argsort(self.gpis)
        idx_gpi_other = np.argsort(other.gpis)
        cellsame = np.all(self.arrcell[idx_gpi]
                          == other.arrcell[idx_gpi_other])
        return np.all([basicsame, cellsame])


def lonlat2cell(lon, lat, cellsize=5., cellsize_lon=None, cellsize_lat=None):
    """
    Partition lon, lat points into cells.

    Parameters
    ----------
    lat: float64, or numpy.ndarray
        Latitude.
    lon: float64, or numpy.ndarray
        Longitude.
    cellsize: float
        Cell size in degrees.
    cellsize_lon : float, optional
        Cell size in degrees on the longitude axis.
    cellsize_lat : float, optional
        Cell size in degrees on the latitude axis.

    Returns
    -------
    cell: int32, or numpy.ndarray
        Cell numbers.
    """
    if cellsize_lon is None:
        cellsize_lon = cellsize
    if cellsize_lat is None:
        cellsize_lat = cellsize
    y = np.clip(np.floor((np.double(lat) +
                          (np.double(90.0) + 1e-9)) / cellsize_lat), 0, 180)
    x = np.clip(np.floor((np.double(lon) + (np.double(180.0) + 1e-9))
                         / cellsize_lon), 0, 360)
    cells = np.int32(x * (np.double(180.0) / cellsize_lat) + y)

    max_cells = ((np.double(180.0) / cellsize_lat) *
                 (np.double(360.0)) / cellsize_lon)
    cells = np.where(cells > max_cells - 1, cells - max_cells, cells)
    return np.int32(cells)


def gridfromdims(londim, latdim, **kwargs):
    """
    Defines new grid object from latitude and longitude dimensions. Latitude
    and longitude dimensions are 1D arrays that give the latitude and
    longitude values of a 2D latitude-longitude array.

    Parameters
    ----------
    londim : numpy.ndarray
        longitude dimension
    latdim : numpy.ndarray
        latitude dimension

    Returns
    -------
    grid : BasicGrid
        New grid object.
    """
    lons, lats = np.meshgrid(londim, latdim)
    return BasicGrid(lons.flatten(), lats.flatten(),
                     shape=(len(latdim), len(londim)), **kwargs)


def genreg_grid(grd_spc_lat=1, grd_spc_lon=1, minlat=-90.0, maxlat=90.0,
                minlon=-180.0, maxlon=180.0, **kwargs):
    """
    Define a global regular lon lat grid which starts in the North Western
    Corner of minlon, maxlat. The grid points are defined to be in the middle
    of a grid cell. e.g. the first point on a 1x1 degree grid with
    minlon -180.0 and maxlat 90.0 will be at -179.5 longitude, 89.5 latitude.

    Parameters
    ----------
    grd_spc_lat: float, optional
        Grid spacing in latitude direction.
    grd_spc_lon: float, optional
        Grid spacing in longitude direction.
    minlat : float, optional
        Minimum latitude of the grid.
    maxlat : float, optional
        Maximum latitude of the grid.
    minlon : float, optional
        Minimum longitude of the grid.
    maxlon : float, optional
        Maximum longitude of the grid.
    """
    lon_dim = np.arange(minlon + grd_spc_lon / 2.0, maxlon, grd_spc_lon)
    lat_dim = np.arange(maxlat - grd_spc_lat / 2.0, minlat, -grd_spc_lat)

    return gridfromdims(lon_dim, lat_dim, **kwargs)


def _element_iterable(el):
    """
    Test if a element is iterable

    Parameters
    ----------
    el: object


    Returns
    -------
    iterable: boolean
       if True then then el is iterable
       if Fales then not
    """
    try:
        el[0]
        iterable = True
    except (TypeError, IndexError):
        iterable = False

    return iterable


def reorder_to_cellsize(grid, cellsize_lat, cellsize_lon):
    """
    Reorder grid points in one grid to follow the
    ordering of differently sized cells. This is useful if
    e.g. a 10x10 degree CellGrid should be traversed
    in an order compatible with a 5x5 degree CellGrid.

    Parameters
    ----------
    grid: :py:class:`pygeogrids.grids.CellGrid`
        input grid
    cellsize_lat: float
        cellsize in latitude direction
    cellsize_lon: float
        cellsize in longitude direction

    Returns
    -------
    new_grid: :py:class:`pygeogrids.grids.CellGrid`
        output grid with original cell sizes but
        different ordering.
    """

    cell_grid = grid.to_cell_grid(cellsize_lat=cellsize_lat,
                                  cellsize_lon=cellsize_lon)
    cell_sort = np.argsort(cell_grid.arrcell)
    new_arrlon = grid.arrlon[cell_sort]
    new_arrlat = grid.arrlat[cell_sort]
    new_arrcell = grid.arrcell[cell_sort]
    new_gpis = grid.gpis[cell_sort]
    new_subset = None
    if grid.subset is not None:
        full_subset = np.zeros(new_arrlon.size)
        full_subset[grid.subset] = 1
        new_full_subset = full_subset[cell_sort]
        new_subset = np.where(new_full_subset == 1)[0]
    return CellGrid(new_arrlon, new_arrlat, new_arrcell,
                    gpis=new_gpis,
                    subset=new_subset)



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import numpy as np
import pyproj


class GeodeticDatum():
    """
    Class representing a geodetic datum providing transformation and
    calculation functionality

    Parameters
    ----------
    ellString : string
        String of geodetic datum (ellipsoid) as provided in pyproj

    """

    def __init__(self, ellps, **kwargs):
        kwargs['ellps'] = ellps
        self.geod = pyproj.Geod(**kwargs)
        self.geod.e = np.sqrt(self.geod.es)
        self.name = ellps

    def getParameter(self):
        """
        Method to transform lon/lat to ECEF (Earth-Centered, Earth-Fixed)
        coordinates representing a 3d Cartesian coordinate system.

        Parameters
        ----------
        lon : numpy.array, list or float
            longitudes of the points in the grid
        lat : numpy.array, list or float
            latitudes of the points in the grid

        Returns
        -------
        x, y, z : np.array
            3D cartesian coordinates
        """
        return self.geod.a, self.geod.b, self.geod.f, self.geod.e

    def toECEF(self, lon, lat):
        """
        Method to transform lon/lat to ECEF (Earth-Centered, Earth-Fixed)
        coordinates representing a 3d Cartesian coordinate system.

        Parameters
        ----------
        lon : numpy.array, list or float
            longitudes of the points in the grid
        lat : numpy.array, list or float
            geodatic latitudes of the points in the grid

        Returns
        -------
        x, y, z : np.array
            3D cartesian coordinates
        """
        if _element_iterable(lat) and lat.shape == lon.shape:
            lat = np.array(lat, dtype=np.float64)
            lon = np.array(lon, dtype=np.float64)

        N = self.EllN(lat)
        lon = np.deg2rad(lon)
        lat = np.deg2rad(lat)

        x = N * np.cos(lat) * np.cos(lon)
        y = N * np.cos(lat) * np.sin(lon)
        z = N * (1 - self.geod.es) * np.sin(lat)

        return x, y, z

    def ParallelRadi(self, lat):
        """
        Method to get the radius the parallel at a given latitude.

        Parameters
        ----------
        lat : numpy.array, list or float
            latitudes of the points in the grid

        Returns
        -------
        radius : np.array, float
            Radius of parallel
        """
        x, y, _ = self.toECEF(0., lat)
        return np.sqrt(x ** 2 + y ** 2)

    def GeocentricLat(self, lat):
        """
        Method to calculate the geocentric from the geodatic latitude.

        Parameters
        ----------
        lat : numpy.array, list or float
            Geodatic latitudes of the points in the grid

        Returns
        -------
        lat_geocentric : np.array, float
            Geocentric latitude
        """
        if _element_iterable(lat):
            lat = np.array(lat, dtype=np.float64)

        return np.rad2deg(np.arctan((1 - self.geod.e ** 2) *
                                    np.tan(np.deg2rad(lat))))

    def GeodeticLat(self, lat):
        """
        Method to calculate the geodatic from the geocentric latitude.

        Parameters
        ----------
        lat : numpy.array, list or float
            Geocentric latitudes of the points in the grid

        Returns
        -------
        lat_geodatic : np.array, float
            Geodatic latitude
        """
        if _element_iterable(lat):
            lat = np.array(lat, dtype=np.float64)
        return np.rad2deg(np.arctan(np.tan(np.deg2rad(lat)) /
                                    (1 - self.geod.e ** 2)))

    def ReducedLat(self, lat):
        """
        Method to calculate the reduced from the geodatic latitude.

        Parameters
        ----------
        lat : numpy.array, list or float
            Geodatic latitudes of the points in the grid

        Returns
        -------
        lat_reduced : np.array, float
            Reduced latitude
        """
        if _element_iterable(lat):
            lat = np.array(lat, dtype=np.float64)
        return np.rad2deg(np.arctan(np.sqrt(1 - self.geod.e ** 2) *
                                    np.tan(np.deg2rad(lat))))

    def GeocentricDistance(self, lon, lat):
        """
        Method to calculate the geocentric distance to given points

        Parameters
        ----------
        lon : numpy.array, list or float
            Geodatic longitude of the points in the grid
        lat : numpy.array, list or float
            Geodatic latitudes of the points in the grid

        Returns
        -------
        r : np.array, float
            Geocentric radius
        """
        x, y, z = self.toECEF(lon, lat)
        return np.sqrt(x**2 + y**2 + z**2)

    def EllN(self, lat):
        """
        Method to calculate the radius of the prime vertical

        Parameters
        ----------
        lat : numpy.array, list or float
            Geodatic latitudes of the points in the grid in degrees

        Returns
        -------
        r : np.array, float
            radius of the prime vertical
        """

        if _element_iterable(lat):
            lat = np.array(lat, dtype=np.float64)
        return self.geod.a / np.sqrt(1 - (self.geod.es) *
                                     (np.sin(np.deg2rad(lat))) ** 2)

    def EllM(self, lat):
        """
        Method to calculate the radius of the curvature

        Parameters
        ----------
        lat : numpy.array, list or float
            Geodatic latitudes of the points in the grid

        Returns
        -------
        r : np.array, float
            radius of the curvature
        """
        if _element_iterable(lat):
            lat = np.array(lat, dtype=np.float64)
        return (self.geod.a * (1 - self.geod.es)) / \
               ((1 - self.geod.es) *
                (np.sin(np.deg2rad(lat)) ** 2) ** (3. / 2.))

    def GaussianRadi(self, lat):
        """
        Method to calculate the gaussian radius of the curvature

        Parameters
        ----------
        lat : numpy.array, list or float
            Geodatic latitudes of the points in the grid

        Returns
        -------
        r : np.array, float
            gaussian radius of the curvature
        """
        if _element_iterable(lat):
            lat = np.array(lat, dtype=np.float64)
        N = self.EllN(lat)
        M = self.EllM(lat)
        return np.sqrt(M * N)

    def ParallelArcDist(self, lat, lon1, lon2):
        """
        Method to calculate the distance between two points on a given parallel

        Parameters
        ----------
        lat : float
            Geodatic latitudes of the points in the grid
        lon1 : float
            Longitude of point 1 at the given parallel
        lon2 : float
            Longitude of point 2 at the given parallel

        Returns
        -------
        dist : np.array, float
            Parallel arc distance
        """
        lon1 = np.deg2rad(lon1)
        lon2 = np.deg2rad(lon2)
        return self.EllN(lat) * np.cos(np.deg2rad(lat)) * (lon2 - lon1)

    def MeridianArcDist(self, lat1, lat2):
        """
        Method to calculate the distance between two parallels (meridian arc
        distance)

        Parameters
        ----------
        lat1 : numpy.array, float
            Geodatic latitudes 1
        lat2 : numpy.array, float
            Geodatic latitudes 2

        Returns
        -------
        dist : np.array, float
            Meridian arc distance
        """
        if _element_iterable(lat1) and lat1.shape == lat2.shape:
            lat1 = np.array(lat1, dtype=np.float64)
            lat2 = np.array(lat2, dtype=np.float64)
        fazi, bazi, dist = self.geod.inv(0., lat1, 0., lat2)
        return dist


def _element_iterable(el):
    """
    Test if a element is iterable

    Parameters
    ----------
    el: object


    Returns
    -------
    iterable: boolean
       if True then then el is iterable
       if Fales then not
    """
    try:
        el[0]
        iterable = True
    except (TypeError, IndexError):
        iterable = False

    return iterable



# Copyright (c) 2018, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN, DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import unittest
import numpy as np
import numpy.testing as nptest
from netCDF4 import Dataset

from pygeogrids.geodetic_datum import GeodeticDatum
import pygeogrids.netcdf as grid_nc
import pygeogrids as grids
import tempfile


class Test(unittest.TestCase):

    def setUp(self):
        lat, lon = np.arange(180) - 90, np.arange(360) - 180
        self.lats, self.lons = np.meshgrid(lat, lon)
        self.lats, self.lons = self.lats.flatten(), self.lons.flatten()
        self.cells = grids.lonlat2cell(self.lons, self.lats)
        self.subset = np.sort(np.random.choice(np.arange(self.lats.size),
                                               size=500, replace=False))
        self.basic = grids.BasicGrid(self.lons, self.lats, subset=self.subset,
                                     shape=(180, 360))

        self.basic_shape_gpis = grids.BasicGrid(self.lons, self.lats,
                                                gpis=np.arange(self.lats.size),
                                                subset=self.subset,
                                                shape=(180, 360))
        self.basic_generated = grids.genreg_grid(1, 1)
        self.basic_irregular = grids.BasicGrid(np.random.random(360 * 180) * 360 - 180,
                                               np.random.random(
                                                   360 * 180) * 180 - 90,
                                               subset=self.subset)
        self.cellgrid = grids.CellGrid(self.lons, self.lats, self.cells,
                                       subset=self.subset)

        self.cellgrid_shape = grids.CellGrid(self.lons, self.lats, self.cells,
                                             subset=self.subset,
                                             shape=(180, 360))

        self.testfile = tempfile.NamedTemporaryFile().name

    def tearDown(self):
        pass

    def test_save_lonlat_nc(self):
        grid_nc.save_lonlat(self.testfile,
                            self.lons, self.lats,
                            GeodeticDatum('WGS84'),
                            self.cells,
                            subsets={'subset_flag': {'points': self.subset,
                                                     'value': 1.,
                                                     'meaning': 'test_flag'}},
                            global_attrs={'test': 'test_attribute'})

        with Dataset(self.testfile) as nc_data:
            nptest.assert_array_equal(self.lats, nc_data.variables['lat'][:])
            nptest.assert_array_equal(self.lons, nc_data.variables['lon'][:])
            nptest.assert_array_equal(self.cells, nc_data.variables['cell'][:])
            nptest.assert_array_equal(
                self.subset, np.where(nc_data.variables['subset_flag'][:] == 1)[0])
            assert nc_data.test == 'test_attribute'

    def test_save_basicgrid_generated(self):
        grid_nc.save_grid(self.testfile,
                          self.basic,
                          global_attrs={'test': 'test_attribute'})

        with Dataset(self.testfile) as nc_data:
            nptest.assert_array_equal(np.unique(self.lats)[::-1],
                                      nc_data.variables['lat'][:])
            nptest.assert_array_equal(np.unique(self.lons),
                                      nc_data.variables['lon'][:])

            # subsets have to identify the same gpis in the original grid and
            # the stored one.
            stored_subset = np.where(nc_data.variables['subset_flag'][
                                     :].flatten() == 1)[0]
            nptest.assert_array_equal(sorted(self.basic.gpis[self.subset]),
                                      sorted(nc_data.variables['gpi'][:].flatten()[stored_subset]))
            assert nc_data.test == 'test_attribute'
            assert nc_data.shape[0] == 180
            assert nc_data.shape[1] == 360

    def test_save_basicgrid_irregular_nc(self):
        grid_nc.save_grid(self.testfile,
                          self.basic_irregular,
                          global_attrs={'test': 'test_attribute'})

        with Dataset(self.testfile) as nc_data:
            nptest.assert_array_equal(
                self.basic_irregular.arrlat, nc_data.variables['lat'][:])
            nptest.assert_array_equal(
                self.basic_irregular.arrlon, nc_data.variables['lon'][:])
            nptest.assert_array_equal(self.subset,
                                      np.where(nc_data.variables['subset_flag'][:] == 1)[0])
            assert nc_data.test == 'test_attribute'
            assert nc_data.shape == 64800

    def test_save_cellgrid_nc(self):
        grid_nc.save_grid(self.testfile,
                          self.cellgrid,
                          global_attrs={'test': 'test_attribute'})

        with Dataset(self.testfile) as nc_data:
            nptest.assert_array_equal(self.lats, nc_data.variables['lat'][:])
            nptest.assert_array_equal(self.lons, nc_data.variables['lon'][:])
            nptest.assert_array_equal(self.cells, nc_data.variables['cell'][:])
            nptest.assert_array_equal(
                self.subset, np.where(nc_data.variables['subset_flag'][:] == 1)[0])
            assert nc_data.test == 'test_attribute'

    def test_save_load_basicgrid(self):
        grid_nc.save_grid(self.testfile,
                          self.basic)

        loaded_grid = grid_nc.load_grid(self.testfile)
        assert self.basic == loaded_grid

    def test_save_load_basicgrid_shape_gpis(self):
        grid_nc.save_grid(self.testfile,
                          self.basic_shape_gpis)

        loaded_grid = grid_nc.load_grid(self.testfile)
        assert self.basic_shape_gpis == loaded_grid

    def test_save_load_basicgrid_irregular(self):
        grid_nc.save_grid(self.testfile,
                          self.basic_irregular)

        loaded_grid = grid_nc.load_grid(self.testfile)
        assert self.basic_irregular == loaded_grid

    def test_save_load_cellgrid(self):
        grid_nc.save_grid(self.testfile,
                          self.cellgrid)

        loaded_grid = grid_nc.load_grid(self.testfile)
        assert self.cellgrid == loaded_grid

    def test_save_load_cellgrid_shape(self):
        grid_nc.save_grid(self.testfile,
                          self.cellgrid_shape)

        loaded_grid = grid_nc.load_grid(self.testfile)
        assert self.cellgrid_shape == loaded_grid


def test_store_load_regular_2D_grid_custom_gpis():
    """
    Test the storing/loading of a 2D grid when the gpis are in a custom
    ordering.
    """
    londim = np.arange(-180.0, 180.0, 60)
    latdim = np.arange(-90.0, 90.0, 30)
    lons, lats = np.meshgrid(londim, latdim)
    gpis = np.arange(lons.flatten().size).reshape(lons.shape)
    grid = grids.BasicGrid(lons.flatten(), lats.flatten(),
                           gpis.flatten(), shape=lons.shape)
    testfile = tempfile.NamedTemporaryFile().name
    grid_nc.save_grid(testfile, grid)
    grid_loaded = grid_nc.load_grid(testfile)
    assert grid == grid_loaded


def test_store_load_regular_2D_grid():
    """
    Test the storing/loading of a 2D grid when the gpis are in a custom
    ordering.
    """
    londim = np.arange(-180.0, 180.0, 60)
    latdim = np.arange(90.0, -90.0, -30)
    lons, lats = np.meshgrid(londim, latdim)
    gpis = np.arange(lons.flatten().size).reshape(lons.shape)
    grid = grids.BasicGrid(lons.flatten(), lats.flatten(),
                           gpis.flatten(), shape=lons.shape)
    testfile = tempfile.NamedTemporaryFile().name
    grid_nc.save_grid(testfile, grid)
    grid_loaded = grid_nc.load_grid(testfile)
    assert grid == grid_loaded


def test_sort_lon_lat_for_netcdf_transposed():
    """
    Test the sorting of an array for netcdf storage
    """
    londim = np.arange(-180.0, 180.0, 60)
    latdim = np.arange(90.0, -90.0, -30)
    lats, lons = np.meshgrid(latdim, londim)
    gpis = np.arange(lons.flatten().size).reshape(lons.shape)
    rand_idx = np.random.permutation(gpis.flatten().size)
    lons_rand = lons.flatten()[rand_idx].reshape(lons.shape)
    lats_rand = lats.flatten()[rand_idx].reshape(lats.shape)
    gpis_rand = gpis.flatten()[rand_idx].reshape(gpis.shape)
    lons_sorted, lats_sorted, gpis_sorted = grid_nc.sort_for_netcdf(
        lons_rand, lats_rand, gpis_rand)
    nptest.assert_almost_equal(lons_sorted, lons.T)
    nptest.assert_almost_equal(lats_sorted, lats.T)
    nptest.assert_almost_equal(gpis_sorted, gpis.T)


def test_sort_lon_lat_for_netcdf():
    """
    Test the sorting of an array for netcdf storage
    """
    londim = np.arange(-180.0, 180.0, 60)
    latdim = np.arange(90.0, -90.0, -30)
    lons, lats = np.meshgrid(londim, latdim)
    gpis = np.arange(lons.flatten().size).reshape(lons.shape)
    rand_idx = np.random.permutation(gpis.flatten().size)
    lons_rand = lons.flatten()[rand_idx].reshape(lons.shape)
    lats_rand = lats.flatten()[rand_idx].reshape(lats.shape)
    gpis_rand = gpis.flatten()[rand_idx].reshape(gpis.shape)
    lons_sorted, lats_sorted, gpis_sorted = grid_nc.sort_for_netcdf(
        lons_rand, lats_rand, gpis_rand)
    nptest.assert_almost_equal(lons_sorted, lons)
    nptest.assert_almost_equal(lats_sorted, lats)
    nptest.assert_almost_equal(gpis_sorted, gpis)


if __name__ == "__main__":
    unittest.main()

